# This file was automatically generated by SWIG (http://www.swig.org).
# Version 3.0.12
#
# Do not make changes to this file unless you know what you are doing--modify
# the SWIG interface file instead.

from sys import version_info as _swig_python_version_info
if _swig_python_version_info >= (2, 7, 0):
    def swig_import_helper():
        import importlib
        pkg = __name__.rpartition('.')[0]
        mname = '.'.join((pkg, '_tabu_search')).lstrip('.')
        try:
            return importlib.import_module(mname)
        except ImportError:
            return importlib.import_module('_tabu_search')
    _tabu_search = swig_import_helper()
    del swig_import_helper
elif _swig_python_version_info >= (2, 6, 0):
    def swig_import_helper():
        from os.path import dirname
        import imp
        fp = None
        try:
            fp, pathname, description = imp.find_module('_tabu_search', [dirname(__file__)])
        except ImportError:
            import _tabu_search
            return _tabu_search
        try:
            _mod = imp.load_module('_tabu_search', fp, pathname, description)
        finally:
            if fp is not None:
                fp.close()
        return _mod
    _tabu_search = swig_import_helper()
    del swig_import_helper
else:
    import _tabu_search
del _swig_python_version_info

try:
    _swig_property = property
except NameError:
    pass  # Python < 2.2 doesn't have 'property'.

try:
    import builtins as __builtin__
except ImportError:
    import __builtin__

def _swig_setattr_nondynamic(self, class_type, name, value, static=1):
    if (name == "thisown"):
        return self.this.own(value)
    if (name == "this"):
        if type(value).__name__ == 'SwigPyObject':
            self.__dict__[name] = value
            return
    method = class_type.__swig_setmethods__.get(name, None)
    if method:
        return method(self, value)
    if (not static):
        if _newclass:
            object.__setattr__(self, name, value)
        else:
            self.__dict__[name] = value
    else:
        raise AttributeError("You cannot add attributes to %s" % self)


def _swig_setattr(self, class_type, name, value):
    return _swig_setattr_nondynamic(self, class_type, name, value, 0)


def _swig_getattr(self, class_type, name):
    if (name == "thisown"):
        return self.this.own()
    method = class_type.__swig_getmethods__.get(name, None)
    if method:
        return method(self)
    raise AttributeError("'%s' object has no attribute '%s'" % (class_type.__name__, name))


def _swig_repr(self):
    try:
        strthis = "proxy of " + self.this.__repr__()
    except __builtin__.Exception:
        strthis = ""
    return "<%s.%s; %s >" % (self.__class__.__module__, self.__class__.__name__, strthis,)

try:
    _object = object
    _newclass = 1
except __builtin__.Exception:
    class _object:
        pass
    _newclass = 0

class SwigPyIterator(_object):
    __swig_setmethods__ = {}
    __setattr__ = lambda self, name, value: _swig_setattr(self, SwigPyIterator, name, value)
    __swig_getmethods__ = {}
    __getattr__ = lambda self, name: _swig_getattr(self, SwigPyIterator, name)

    def __init__(self, *args, **kwargs):
        raise AttributeError("No constructor defined - class is abstract")
    __repr__ = _swig_repr
    __swig_destroy__ = _tabu_search.delete_SwigPyIterator
    __del__ = lambda self: None

    def value(self):
        return _tabu_search.SwigPyIterator_value(self)

    def incr(self, n=1):
        return _tabu_search.SwigPyIterator_incr(self, n)

    def decr(self, n=1):
        return _tabu_search.SwigPyIterator_decr(self, n)

    def distance(self, x):
        return _tabu_search.SwigPyIterator_distance(self, x)

    def equal(self, x):
        return _tabu_search.SwigPyIterator_equal(self, x)

    def copy(self):
        return _tabu_search.SwigPyIterator_copy(self)

    def next(self):
        return _tabu_search.SwigPyIterator_next(self)

    def __next__(self):
        return _tabu_search.SwigPyIterator___next__(self)

    def previous(self):
        return _tabu_search.SwigPyIterator_previous(self)

    def advance(self, n):
        return _tabu_search.SwigPyIterator_advance(self, n)

    def __eq__(self, x):
        return _tabu_search.SwigPyIterator___eq__(self, x)

    def __ne__(self, x):
        return _tabu_search.SwigPyIterator___ne__(self, x)

    def __iadd__(self, n):
        return _tabu_search.SwigPyIterator___iadd__(self, n)

    def __isub__(self, n):
        return _tabu_search.SwigPyIterator___isub__(self, n)

    def __add__(self, n):
        return _tabu_search.SwigPyIterator___add__(self, n)

    def __sub__(self, *args):
        return _tabu_search.SwigPyIterator___sub__(self, *args)
    def __iter__(self):
        return self
SwigPyIterator_swigregister = _tabu_search.SwigPyIterator_swigregister
SwigPyIterator_swigregister(SwigPyIterator)

class IntVector(_object):
    __swig_setmethods__ = {}
    __setattr__ = lambda self, name, value: _swig_setattr(self, IntVector, name, value)
    __swig_getmethods__ = {}
    __getattr__ = lambda self, name: _swig_getattr(self, IntVector, name)
    __repr__ = _swig_repr

    def iterator(self):
        return _tabu_search.IntVector_iterator(self)
    def __iter__(self):
        return self.iterator()

    def __nonzero__(self):
        return _tabu_search.IntVector___nonzero__(self)

    def __bool__(self):
        return _tabu_search.IntVector___bool__(self)

    def __len__(self):
        return _tabu_search.IntVector___len__(self)

    def __getslice__(self, i, j):
        return _tabu_search.IntVector___getslice__(self, i, j)

    def __setslice__(self, *args):
        return _tabu_search.IntVector___setslice__(self, *args)

    def __delslice__(self, i, j):
        return _tabu_search.IntVector___delslice__(self, i, j)

    def __delitem__(self, *args):
        return _tabu_search.IntVector___delitem__(self, *args)

    def __getitem__(self, *args):
        return _tabu_search.IntVector___getitem__(self, *args)

    def __setitem__(self, *args):
        return _tabu_search.IntVector___setitem__(self, *args)

    def pop(self):
        return _tabu_search.IntVector_pop(self)

    def append(self, x):
        return _tabu_search.IntVector_append(self, x)

    def empty(self):
        return _tabu_search.IntVector_empty(self)

    def size(self):
        return _tabu_search.IntVector_size(self)

    def swap(self, v):
        return _tabu_search.IntVector_swap(self, v)

    def begin(self):
        return _tabu_search.IntVector_begin(self)

    def end(self):
        return _tabu_search.IntVector_end(self)

    def rbegin(self):
        return _tabu_search.IntVector_rbegin(self)

    def rend(self):
        return _tabu_search.IntVector_rend(self)

    def clear(self):
        return _tabu_search.IntVector_clear(self)

    def get_allocator(self):
        return _tabu_search.IntVector_get_allocator(self)

    def pop_back(self):
        return _tabu_search.IntVector_pop_back(self)

    def erase(self, *args):
        return _tabu_search.IntVector_erase(self, *args)

    def __init__(self, *args):
        this = _tabu_search.new_IntVector(*args)
        try:
            self.this.append(this)
        except __builtin__.Exception:
            self.this = this

    def push_back(self, x):
        return _tabu_search.IntVector_push_back(self, x)

    def front(self):
        return _tabu_search.IntVector_front(self)

    def back(self):
        return _tabu_search.IntVector_back(self)

    def assign(self, n, x):
        return _tabu_search.IntVector_assign(self, n, x)

    def resize(self, *args):
        return _tabu_search.IntVector_resize(self, *args)

    def insert(self, *args):
        return _tabu_search.IntVector_insert(self, *args)

    def reserve(self, n):
        return _tabu_search.IntVector_reserve(self, n)

    def capacity(self):
        return _tabu_search.IntVector_capacity(self)
    __swig_destroy__ = _tabu_search.delete_IntVector
    __del__ = lambda self: None
IntVector_swigregister = _tabu_search.IntVector_swigregister
IntVector_swigregister(IntVector)

class DoubleVector(_object):
    __swig_setmethods__ = {}
    __setattr__ = lambda self, name, value: _swig_setattr(self, DoubleVector, name, value)
    __swig_getmethods__ = {}
    __getattr__ = lambda self, name: _swig_getattr(self, DoubleVector, name)
    __repr__ = _swig_repr

    def iterator(self):
        return _tabu_search.DoubleVector_iterator(self)
    def __iter__(self):
        return self.iterator()

    def __nonzero__(self):
        return _tabu_search.DoubleVector___nonzero__(self)

    def __bool__(self):
        return _tabu_search.DoubleVector___bool__(self)

    def __len__(self):
        return _tabu_search.DoubleVector___len__(self)

    def __getslice__(self, i, j):
        return _tabu_search.DoubleVector___getslice__(self, i, j)

    def __setslice__(self, *args):
        return _tabu_search.DoubleVector___setslice__(self, *args)

    def __delslice__(self, i, j):
        return _tabu_search.DoubleVector___delslice__(self, i, j)

    def __delitem__(self, *args):
        return _tabu_search.DoubleVector___delitem__(self, *args)

    def __getitem__(self, *args):
        return _tabu_search.DoubleVector___getitem__(self, *args)

    def __setitem__(self, *args):
        return _tabu_search.DoubleVector___setitem__(self, *args)

    def pop(self):
        return _tabu_search.DoubleVector_pop(self)

    def append(self, x):
        return _tabu_search.DoubleVector_append(self, x)

    def empty(self):
        return _tabu_search.DoubleVector_empty(self)

    def size(self):
        return _tabu_search.DoubleVector_size(self)

    def swap(self, v):
        return _tabu_search.DoubleVector_swap(self, v)

    def begin(self):
        return _tabu_search.DoubleVector_begin(self)

    def end(self):
        return _tabu_search.DoubleVector_end(self)

    def rbegin(self):
        return _tabu_search.DoubleVector_rbegin(self)

    def rend(self):
        return _tabu_search.DoubleVector_rend(self)

    def clear(self):
        return _tabu_search.DoubleVector_clear(self)

    def get_allocator(self):
        return _tabu_search.DoubleVector_get_allocator(self)

    def pop_back(self):
        return _tabu_search.DoubleVector_pop_back(self)

    def erase(self, *args):
        return _tabu_search.DoubleVector_erase(self, *args)

    def __init__(self, *args):
        this = _tabu_search.new_DoubleVector(*args)
        try:
            self.this.append(this)
        except __builtin__.Exception:
            self.this = this

    def push_back(self, x):
        return _tabu_search.DoubleVector_push_back(self, x)

    def front(self):
        return _tabu_search.DoubleVector_front(self)

    def back(self):
        return _tabu_search.DoubleVector_back(self)

    def assign(self, n, x):
        return _tabu_search.DoubleVector_assign(self, n, x)

    def resize(self, *args):
        return _tabu_search.DoubleVector_resize(self, *args)

    def insert(self, *args):
        return _tabu_search.DoubleVector_insert(self, *args)

    def reserve(self, n):
        return _tabu_search.DoubleVector_reserve(self, n)

    def capacity(self):
        return _tabu_search.DoubleVector_capacity(self)
    __swig_destroy__ = _tabu_search.delete_DoubleVector
    __del__ = lambda self: None
DoubleVector_swigregister = _tabu_search.DoubleVector_swigregister
DoubleVector_swigregister(DoubleVector)

class IntArray(_object):
    __swig_setmethods__ = {}
    __setattr__ = lambda self, name, value: _swig_setattr(self, IntArray, name, value)
    __swig_getmethods__ = {}
    __getattr__ = lambda self, name: _swig_getattr(self, IntArray, name)
    __repr__ = _swig_repr

    def iterator(self):
        return _tabu_search.IntArray_iterator(self)
    def __iter__(self):
        return self.iterator()

    def __nonzero__(self):
        return _tabu_search.IntArray___nonzero__(self)

    def __bool__(self):
        return _tabu_search.IntArray___bool__(self)

    def __len__(self):
        return _tabu_search.IntArray___len__(self)

    def __getslice__(self, i, j):
        return _tabu_search.IntArray___getslice__(self, i, j)

    def __setslice__(self, *args):
        return _tabu_search.IntArray___setslice__(self, *args)

    def __delslice__(self, i, j):
        return _tabu_search.IntArray___delslice__(self, i, j)

    def __delitem__(self, *args):
        return _tabu_search.IntArray___delitem__(self, *args)

    def __getitem__(self, *args):
        return _tabu_search.IntArray___getitem__(self, *args)

    def __setitem__(self, *args):
        return _tabu_search.IntArray___setitem__(self, *args)

    def pop(self):
        return _tabu_search.IntArray_pop(self)

    def append(self, x):
        return _tabu_search.IntArray_append(self, x)

    def empty(self):
        return _tabu_search.IntArray_empty(self)

    def size(self):
        return _tabu_search.IntArray_size(self)

    def swap(self, v):
        return _tabu_search.IntArray_swap(self, v)

    def begin(self):
        return _tabu_search.IntArray_begin(self)

    def end(self):
        return _tabu_search.IntArray_end(self)

    def rbegin(self):
        return _tabu_search.IntArray_rbegin(self)

    def rend(self):
        return _tabu_search.IntArray_rend(self)

    def clear(self):
        return _tabu_search.IntArray_clear(self)

    def get_allocator(self):
        return _tabu_search.IntArray_get_allocator(self)

    def pop_back(self):
        return _tabu_search.IntArray_pop_back(self)

    def erase(self, *args):
        return _tabu_search.IntArray_erase(self, *args)

    def __init__(self, *args):
        this = _tabu_search.new_IntArray(*args)
        try:
            self.this.append(this)
        except __builtin__.Exception:
            self.this = this

    def push_back(self, x):
        return _tabu_search.IntArray_push_back(self, x)

    def front(self):
        return _tabu_search.IntArray_front(self)

    def back(self):
        return _tabu_search.IntArray_back(self)

    def assign(self, n, x):
        return _tabu_search.IntArray_assign(self, n, x)

    def resize(self, *args):
        return _tabu_search.IntArray_resize(self, *args)

    def insert(self, *args):
        return _tabu_search.IntArray_insert(self, *args)

    def reserve(self, n):
        return _tabu_search.IntArray_reserve(self, n)

    def capacity(self):
        return _tabu_search.IntArray_capacity(self)
    __swig_destroy__ = _tabu_search.delete_IntArray
    __del__ = lambda self: None
IntArray_swigregister = _tabu_search.IntArray_swigregister
IntArray_swigregister(IntArray)

class DoubleArray(_object):
    __swig_setmethods__ = {}
    __setattr__ = lambda self, name, value: _swig_setattr(self, DoubleArray, name, value)
    __swig_getmethods__ = {}
    __getattr__ = lambda self, name: _swig_getattr(self, DoubleArray, name)
    __repr__ = _swig_repr

    def iterator(self):
        return _tabu_search.DoubleArray_iterator(self)
    def __iter__(self):
        return self.iterator()

    def __nonzero__(self):
        return _tabu_search.DoubleArray___nonzero__(self)

    def __bool__(self):
        return _tabu_search.DoubleArray___bool__(self)

    def __len__(self):
        return _tabu_search.DoubleArray___len__(self)

    def __getslice__(self, i, j):
        return _tabu_search.DoubleArray___getslice__(self, i, j)

    def __setslice__(self, *args):
        return _tabu_search.DoubleArray___setslice__(self, *args)

    def __delslice__(self, i, j):
        return _tabu_search.DoubleArray___delslice__(self, i, j)

    def __delitem__(self, *args):
        return _tabu_search.DoubleArray___delitem__(self, *args)

    def __getitem__(self, *args):
        return _tabu_search.DoubleArray___getitem__(self, *args)

    def __setitem__(self, *args):
        return _tabu_search.DoubleArray___setitem__(self, *args)

    def pop(self):
        return _tabu_search.DoubleArray_pop(self)

    def append(self, x):
        return _tabu_search.DoubleArray_append(self, x)

    def empty(self):
        return _tabu_search.DoubleArray_empty(self)

    def size(self):
        return _tabu_search.DoubleArray_size(self)

    def swap(self, v):
        return _tabu_search.DoubleArray_swap(self, v)

    def begin(self):
        return _tabu_search.DoubleArray_begin(self)

    def end(self):
        return _tabu_search.DoubleArray_end(self)

    def rbegin(self):
        return _tabu_search.DoubleArray_rbegin(self)

    def rend(self):
        return _tabu_search.DoubleArray_rend(self)

    def clear(self):
        return _tabu_search.DoubleArray_clear(self)

    def get_allocator(self):
        return _tabu_search.DoubleArray_get_allocator(self)

    def pop_back(self):
        return _tabu_search.DoubleArray_pop_back(self)

    def erase(self, *args):
        return _tabu_search.DoubleArray_erase(self, *args)

    def __init__(self, *args):
        this = _tabu_search.new_DoubleArray(*args)
        try:
            self.this.append(this)
        except __builtin__.Exception:
            self.this = this

    def push_back(self, x):
        return _tabu_search.DoubleArray_push_back(self, x)

    def front(self):
        return _tabu_search.DoubleArray_front(self)

    def back(self):
        return _tabu_search.DoubleArray_back(self)

    def assign(self, n, x):
        return _tabu_search.DoubleArray_assign(self, n, x)

    def resize(self, *args):
        return _tabu_search.DoubleArray_resize(self, *args)

    def insert(self, *args):
        return _tabu_search.DoubleArray_insert(self, *args)

    def reserve(self, n):
        return _tabu_search.DoubleArray_reserve(self, n)

    def capacity(self):
        return _tabu_search.DoubleArray_capacity(self)
    __swig_destroy__ = _tabu_search.delete_DoubleArray
    __del__ = lambda self: None
DoubleArray_swigregister = _tabu_search.DoubleArray_swigregister
DoubleArray_swigregister(DoubleArray)

class TabuSearch(_object):
    """
    Tabu Search

    handler = TabuSearch(q, init_solution, tenure, scaleFactor, timeout)

        Args:
            q: QUBO as a list of list, or numpy matrix of double (float64) values.
               q must be symmetric.
            init_solution: List of 0/1 values, which defines the initial state of each variable.
            tenure: Tabu tenure. min(20, num_vars / 4) seems to be a good choice.
            scaleFactor: Scaling factor for elements of q. The elements of q are stored as long ints using
                         internal_q = long int (q * scaleFactor).
            timeout: Total running time in milliseconds.

    energy = handler.bestEnergy()

    solution = handler.bestSolution()
    """

    __swig_setmethods__ = {}
    __setattr__ = lambda self, name, value: _swig_setattr(self, TabuSearch, name, value)
    __swig_getmethods__ = {}
    __getattr__ = lambda self, name: _swig_getattr(self, TabuSearch, name)
    __repr__ = _swig_repr

    def __init__(self, Q, initSol, tenure, scaleFactor, timeout):
        """
        Tabu Search

        handler = TabuSearch(q, init_solution, tenure, scaleFactor, timeout)

            Args:
                q: QUBO as a list of list, or numpy matrix of double (float64) values.
                   q must be symmetric.
                init_solution: List of 0/1 values, which defines the initial state of each variable.
                tenure: Tabu tenure. min(20, num_vars / 4) seems to be a good choice.
                scaleFactor: Scaling factor for elements of q. The elements of q are stored as long ints using
                             internal_q = long int (q * scaleFactor).
                timeout: Total running time in milliseconds.

        energy = handler.bestEnergy()

        solution = handler.bestSolution()
        """
        this = _tabu_search.new_TabuSearch(Q, initSol, tenure, scaleFactor, timeout)
        try:
            self.this.append(this)
        except __builtin__.Exception:
            self.this = this

    def bestEnergy(self):
        return _tabu_search.TabuSearch_bestEnergy(self)

    def bestSolution(self):
        return _tabu_search.TabuSearch_bestSolution(self)
    __swig_destroy__ = _tabu_search.delete_TabuSearch
    __del__ = lambda self: None
TabuSearch_swigregister = _tabu_search.TabuSearch_swigregister
TabuSearch_swigregister(TabuSearch)

class Exception(_object):
    __swig_setmethods__ = {}
    __setattr__ = lambda self, name, value: _swig_setattr(self, Exception, name, value)
    __swig_getmethods__ = {}
    __getattr__ = lambda self, name: _swig_getattr(self, Exception, name)
    __repr__ = _swig_repr

    def __init__(self, error):
        this = _tabu_search.new_Exception(error)
        try:
            self.this.append(this)
        except __builtin__.Exception:
            self.this = this
    __swig_destroy__ = _tabu_search.delete_Exception
    __del__ = lambda self: None

    def what(self):
        return _tabu_search.Exception_what(self)
Exception_swigregister = _tabu_search.Exception_swigregister
Exception_swigregister(Exception)

# This file is compatible with both classic and new-style classes.





# Copyright 2018 D-Wave Systems Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""A dimod sampler_ that uses the MST2 multistart tabu search algorithm."""

import random

import numpy
import dimod

from tabu import TabuSearch


class TabuSampler(dimod.Sampler):
    """A tabu-search sampler.

    Examples:
        This example solves a two-variable Ising model.

        >>> from tabu import TabuSampler
        >>> samples = TabuSampler().sample_ising({'a': -0.5, 'b': 1.0}, {'ab': -1})
        >>> list(samples.data()) # doctest: +SKIP
        [Sample(sample={'a': -1, 'b': -1}, energy=-1.5, num_occurrences=1)]
        >>> samples.first.energy
        -1.5

    """

    properties = None
    parameters = None

    def __init__(self):
        self.parameters = {'tenure': [],
                           'scale_factor': [],
                           'timeout': [],
                           'num_reads': [],
                           'init_solution': []}
        self.properties = {}

    def sample(self, bqm, init_solution=None, tenure=None, scale_factor=1, timeout=20, num_reads=1):
        """Run a tabu search on a given binary quadratic model.

        Args:
            bqm (:obj:`~dimod.BinaryQuadraticModel`):
                The binary quadratic model (BQM) to be sampled.
            init_solution (:obj:`~dimod.SampleSet`, optional):
                Single sample that sets an initial state for all the problem variables.
                Default is a random initial state.
            tenure (int, optional):
                Tabu tenure, which is the length of the tabu list, or number of recently
                explored solutions kept in memory.
                Default is a quarter of the number of problem variables up to
                a maximum value of 20.
            scale_factor (number, optional):
                Scaling factor for linear and quadratic biases in the BQM. Internally, the BQM is
                converted to a QUBO matrix, and elements are stored as long ints
                using ``internal_q = long int (q * scale_factor)``.
            timeout (int, optional):
                Total running time in milliseconds.
            num_reads (int, optional): Number of reads. Each run of the tabu algorithm
                generates a sample.

        Returns:
            :obj:`~dimod.SampleSet`: A `dimod` :obj:`.~dimod.SampleSet` object.

        Examples:
            This example provides samples for a two-variable QUBO model.

            >>> from tabu import TabuSampler
            >>> import dimod
            >>> sampler = TabuSampler()
            >>> Q = {(0, 0): -1, (1, 1): -1, (0, 1): 2}
            >>> bqm = dimod.BinaryQuadraticModel.from_qubo(Q, offset=0.0)
            >>> samples = sampler.sample(bqm)
            >>> samples.record[0].energy
            -1.0

        """

        # input checking and defaults calculation
        # TODO: one "read" per sample in init_solution sampleset
        if init_solution is not None:
            if not isinstance(init_solution, dimod.SampleSet):
                raise TypeError("'init_solution' should be a 'dimod.SampleSet' instance")
            if len(init_solution.record) < 1:
                raise ValueError("'init_solution' should contain at least one sample")
            if len(init_solution.record[0].sample) != len(bqm):
                raise ValueError("'init_solution' sample dimension different from BQM")
            init_sample = self._bqm_sample_to_tabu_sample(
                init_solution.change_vartype(dimod.BINARY, inplace=False).record[0].sample, bqm.binary)
        else:
            init_sample = None

        if not bqm:
            return dimod.SampleSet.from_samples([], energy=0, vartype=bqm.vartype)

        if tenure is None:
            tenure = int(max(min(20, len(bqm) / 4), 1))
        if not isinstance(tenure, int):
            raise TypeError("'tenure' should be an integer in range [1, num_vars - 1]")
        if not 0 < tenure < len(bqm):
            raise ValueError("'tenure' should be an integer in range [1, num_vars - 1]")

        if not isinstance(num_reads, int):
            raise TypeError("'num_reads' should be a positive integer")
        if num_reads < 1:
            raise ValueError("'num_reads' should be a positive integer")

        qubo = self._bqm_to_tabu_qubo(bqm.binary)

        # run Tabu search
        samples = []
        energies = []
        for _ in range(num_reads):
            if init_sample is None:
                init_sample = self._bqm_sample_to_tabu_sample(self._random_sample(bqm.binary), bqm.binary)
            r = TabuSearch(qubo, init_sample, tenure, scale_factor, timeout)
            sample = self._tabu_sample_to_bqm_sample(list(r.bestSolution()), bqm.binary)
            energy = bqm.binary.energy(sample)
            samples.append(sample)
            energies.append(energy)

        response = dimod.SampleSet.from_samples(
            samples, energy=energies, vartype=dimod.BINARY)
        response.change_vartype(bqm.vartype, inplace=True)
        return response

    def _bqm_to_tabu_qubo(self, bqm):
        # Note: normally, conversion would be: `ud + ud.T - numpy.diag(numpy.diag(ud))`,
        # but the Tabu solver we're using requires slightly different qubo matrix.
        varorder = sorted(list(bqm.adj.keys()))
        ud = 0.5 * bqm.to_numpy_matrix(varorder)
        symm = ud + ud.T
        qubo = symm.tolist()
        return qubo

    def _bqm_sample_to_tabu_sample(self, sample, bqm):
        assert len(sample) == len(bqm)
        _, values = zip(*sorted(self._sample_as_dict(sample).items()))
        return list(map(int, values))

    def _tabu_sample_to_bqm_sample(self, sample, bqm):
        varorder = sorted(list(bqm.adj.keys()))
        assert len(sample) == len(varorder)
        return dict(zip(varorder, sample))

    def _sample_as_dict(self, sample):
        """Convert list-like ``sample`` (list/dict/dimod.SampleView),
        ``list: var``, to ``map: idx -> var``.
        """
        if isinstance(sample, dict):
            return sample
        if isinstance(sample, (list, numpy.ndarray)):
            sample = enumerate(sample)
        return dict(sample)

    def _random_sample(self, bqm):
        values = list(bqm.vartype.value)
        return {i: random.choice(values) for i in bqm.variables}


if __name__ == "__main__":
    from pprint import pprint

    print("TabuSampler:")
    bqm = dimod.BinaryQuadraticModel(
        {'a': 0.0, 'b': -1.0, 'c': 0.5},
        {('a', 'b'): -1.0, ('b', 'c'): 1.5},
        offset=0.0, vartype=dimod.BINARY)
    response = TabuSampler().sample(bqm, num_reads=10)
    pprint(list(response.data()))

    print("ExactSolver:")
    response = dimod.ExactSolver().sample(bqm)
    pprint(list(response.data(sorted_by='energy')))
